"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MI2 = exports.escape = void 0;
const backend_1 = require("../backend");
const ChildProcess = require("child_process");
const events_1 = require("events");
const mi_parse_1 = require("../mi_parse");
const linuxTerm = require("../linux/console");
const net = require("net");
const fs = require("fs");
const path_1 = require("path");
const nativePath = require("path");
const path = path_1.posix;
const ssh2_1 = require("ssh2");
function escape(str) {
    return str.replace(/\\/g, "\\\\").replace(/"/g, "\\\"");
}
exports.escape = escape;
function pausecomp(millis) {
    var date;
    var curDate;
    date = new Date();
    do {
        curDate = new Date();
    } while (curDate - date < millis);
}
//const nonOutput = /^(?:\d*|undefined)[\*\+\=]|[\~\@\&\^]/;
const nonOutput = /^(?:\d*|undefined)[\*\=]|[\~\@\&\^]/; // Dubner removed '+'
const gdbMatch = /(?:\d*|undefined)\(gdb\)/;
const numRegex = /\d+/;
//function couldBeOutput(line: string) {
//	if (nonOutput.exec(line))
//		return false;
//	return true;
//}
function couldBeOutput(line) {
    // Return false if line starts with GDB/MI keyword
    const gdbmi = [
        "(gdb)",
        // Result records
        "^done",
        "^running",
        "^connected",
        "^error",
        "^exit",
        // Stream records
        "~\"",
        "@\"",
        "&\"",
        // Async records
        "*running",
        "*stopped",
        "=thread-",
        "=library-",
        "=traceframe-",
        "=tsv-",
        "=tsv-",
        "=breakpoint-",
        "=record-",
        "=cmd-",
        "=memory-",
    ];
    // Skip leading digits:
    var nstart = 0;
    var patt = new RegExp("^[0-9]");
    while (nstart < line.length) {
        var onechar = line.slice(nstart, nstart + 1);
        if (!patt.test(onechar)) {
            break;
        }
        nstart += 1;
    }
    for (const possible of gdbmi) {
        var chunk = line.slice(nstart, nstart + possible.length);
        if (possible == chunk) {
            return false;
        }
    }
    return true;
}
const trace = false;
class MI2 extends events_1.EventEmitter {
    constructor(application, preargs, extraargs, procEnv) {
        super();
        this.application = application;
        this.preargs = preargs;
        this.extraargs = extraargs;
        this.prettyPrint = true;
        this.currentToken = 1;
        this.handlers = {};
        this.breakpoints = new Map();
        this.starState = ""; // Current MI *running, *stopped, and so on state.
        if (procEnv) {
            const env = {};
            // Duplicate process.env so we don't override it
            for (const key in process.env)
                if (process.env.hasOwnProperty(key))
                    env[key] = process.env[key];
            // Overwrite with user specified variables
            for (const key in procEnv) {
                if (procEnv.hasOwnProperty(key)) {
                    if (procEnv === null)
                        delete env[key];
                    else
                        env[key] = procEnv[key];
                }
            }
            this.procEnv = env;
        }
    }
    load(cwd, program, procArgs, separateConsole) {
        if (!nativePath.isAbsolute(program))
            program = nativePath.join(cwd, program);
        return new Promise((resolve, reject) => {
            this.isSSH = false;
            const args = this.preargs.concat(this.extraargs || []);
            this.process = ChildProcess.spawn(this.application, args, { cwd: cwd, env: this.procEnv });
            this.process.stdout.on("data", this.stdout.bind(this));
            this.process.stderr.on("data", this.stderr.bind(this));
            this.process.on("exit", (() => { this.emit("quit"); }).bind(this));
            this.process.on("error", ((err) => { this.emit("launcherror", err); }).bind(this));
            const promises = this.initCommands(program, cwd);
            if (procArgs && procArgs.length)
                promises.push(this.sendCommand("exec-arguments " + procArgs));
            if (process.platform == "win32") {
                if (separateConsole !== undefined)
                    promises.push(this.sendCommand("gdb-set new-console on"));
                Promise.all(promises).then(() => {
                    this.emit("debug-ready");
                    resolve(true); // Dubner added true
                }, reject);
            }
            else {
                if (separateConsole !== undefined) {
                    linuxTerm.spawnTerminalEmulator(separateConsole).then(tty => {
                        promises.push(this.sendCommand("inferior-tty-set " + tty));
                        Promise.all(promises).then(() => {
                            this.emit("debug-ready");
                            resolve(true); // Dubner added true
                        }, reject);
                    });
                }
                else {
                    Promise.all(promises).then(() => {
                        this.emit("debug-ready");
                        resolve(true); // Dubner added true
                    }, reject);
                }
            }
        });
    }
    ssh(args, cwd, program, procArgs, separateConsole, attach) {
        return new Promise((resolve, reject) => {
            this.isSSH = true;
            this.sshReady = false;
            this.sshConn = new ssh2_1.Client();
            if (separateConsole !== undefined)
                this.log("stderr", "WARNING: Output to terminal emulators are not supported over SSH");
            if (args.forwardX11) {
                this.sshConn.on("x11", (info, accept, reject) => {
                    const xserversock = new net.Socket();
                    xserversock.on("error", (err) => {
                        this.log("stderr", "Could not connect to local X11 server! Did you enable it in your display manager?\n" + err);
                    });
                    xserversock.on("connect", () => {
                        const xclientsock = accept();
                        xclientsock.pipe(xserversock).pipe(xclientsock);
                    });
                    xserversock.connect(args.x11port, args.x11host);
                });
            }
            const connectionArgs = {
                host: args.host,
                port: args.port,
                username: args.user
            };
            if (args.useAgent) {
                connectionArgs.agent = process.env.SSH_AUTH_SOCK;
            }
            else if (args.keyfile) {
                if (require("fs").existsSync(args.keyfile))
                    connectionArgs.privateKey = require("fs").readFileSync(args.keyfile);
                else {
                    this.log("stderr", "SSH key file does not exist!");
                    this.emit("quit");
                    reject();
                    return;
                }
            }
            else {
                connectionArgs.password = args.password;
            }
            this.sshConn.on("ready", () => {
                this.log("stdout", "Running " + this.application + " over ssh...");
                const execArgs = {};
                if (args.forwardX11) {
                    execArgs.x11 = {
                        single: false,
                        screen: args.remotex11screen
                    };
                }
                let sshCMD = this.application + " " + this.preargs.join(" ");
                if (args.bootstrap)
                    sshCMD = args.bootstrap + " && " + sshCMD;
                if (attach)
                    sshCMD += " -p " + program;
                this.sshConn.exec(sshCMD, execArgs, (err, stream) => {
                    if (err) {
                        this.log("stderr", "Could not run " + this.application + " over ssh!");
                        this.log("stderr", err.toString());
                        this.emit("quit");
                        reject();
                        return;
                    }
                    this.sshReady = true;
                    this.stream = stream;
                    stream.on("data", this.stdout.bind(this));
                    stream.stderr.on("data", this.stderr.bind(this));
                    stream.on("exit", (() => {
                        this.emit("quit");
                        this.sshConn.end();
                    }).bind(this));
                    const promises = this.initCommands(program, cwd, true, attach);
                    promises.push(this.sendCommand("environment-cd \"" + escape(cwd) + "\""));
                    if (procArgs && procArgs.length && !attach)
                        promises.push(this.sendCommand("exec-arguments " + procArgs));
                    Promise.all(promises).then(() => {
                        this.emit("debug-ready");
                        resolve(true); // Dubner added true
                    }, reject);
                });
            }).on("error", (err) => {
                this.log("stderr", "Could not run " + this.application + " over ssh!");
                this.log("stderr", err.toString());
                this.emit("quit");
                reject();
            }).connect(connectionArgs);
        });
    }
    initCommands(target, cwd, ssh = false, attach = false) {
        if (ssh) {
            if (!path.isAbsolute(target))
                target = path.join(cwd, target);
        }
        else {
            if (!nativePath.isAbsolute(target))
                target = nativePath.join(cwd, target);
        }
        const cmds = [
            this.sendCommand("gdb-set target-async on", true),
            this.sendCommand("environment-directory \"" + escape(cwd) + "\"", true)
        ];
        if (!attach)
            cmds.push(this.sendCommand("file-exec-and-symbols \"" + escape(target) + "\""));
        if (this.prettyPrint)
            cmds.push(this.sendCommand("enable-pretty-printing"));
        return cmds;
    }
    attach(cwd, executable, target, iexclause, ixscript, primdebug) {
        return new Promise((resolve, reject) => {
            let args = [];
            args = args.concat(this.preargs);
            if (executable && !nativePath.isAbsolute(executable)) {
                executable = nativePath.join(cwd, executable);
            }
            if (!executable && target) {
                executable = "-p";
            }
            let isExtendedRemote = false;
            if (target && target.startsWith("extended-remote")) {
                isExtendedRemote = true;
                args = this.preargs;
            }
            else {
                if (iexclause != undefined) {
                    args = args.concat("-iex");
                    args = args.concat("set 'solib-search-path " + iexclause + "'");
                }
                if (ixscript) {
                    args = args.concat("-ix");
                    args = args.concat(ixscript);
                }
                if (primdebug) {
                    //  One of the numerous possibilities is that the user has supplied us
                    //  with a path to a file, or fifo, containing a single line:
                    //      PID appname\n
                    //
                    //  We need to read that file, and turn it into -ex commands:
                    //
                    // Get the contents of the file
                    try {
                        var primdebugstring = fs.readFileSync(primdebug, "UTF-8");
                        // Strip off the newline at the end
                        primdebugstring = primdebugstring.slice(0, -1);
                        var tokens = primdebugstring.split("\t");
                        if (tokens.length == 2) {
                            // Put in the -p PID
                            args = args.concat("-p");
                            args = args.concat(tokens[0]);
                            // Put in the -ex 'b app_'
                            args = args.concat("-ex");
                            args = args.concat("b " + tokens[1] + "_");
                            // Put in the -ex 'signal SIGCONT'
                            args = args.concat("-ex");
                            args = args.concat("signal SIGCONT");
                        }
                    }
                    catch (error) {
                    }
                }
                if (executable) {
                    args = args.concat(executable);
                }
                if (target) {
                    args = args.concat(target);
                }
            }
            this.process = ChildProcess.spawn(this.application, args, { cwd: cwd, env: this.procEnv });
            this.process.stdout.on("data", this.stdout.bind(this));
            this.process.stderr.on("data", this.stderr.bind(this));
            this.process.on("exit", (() => { this.emit("quit"); }).bind(this));
            this.process.on("error", ((err) => { this.emit("launcherror", err); }).bind(this));
            const commands = [
                //this.sendCommand("gdb-set target-async on"), // Dubner eliminated this to prevent crashing
                this.sendCommand("environment-directory \"" + escape(cwd) + "\"")
            ];
            if (isExtendedRemote) {
                commands.push(this.sendCommand("target-select " + target));
                commands.push(this.sendCommand("file-symbol-file \"" + escape(executable) + "\""));
            }
            Promise.all(commands).then(() => {
                this.emit("debug-ready");
                resolve(true); // Dubner added true
            }, reject);
        });
    }
    connect(cwd, executable, target) {
        return new Promise((resolve, reject) => {
            let args = [];
            if (executable && !nativePath.isAbsolute(executable))
                executable = nativePath.join(cwd, executable);
            if (executable)
                args = args.concat([executable], this.preargs);
            else
                args = this.preargs;
            this.process = ChildProcess.spawn(this.application, args, { cwd: cwd, env: this.procEnv });
            this.process.stdout.on("data", this.stdout.bind(this));
            this.process.stderr.on("data", this.stderr.bind(this));
            this.process.on("exit", (() => { this.emit("quit"); }).bind(this));
            this.process.on("error", ((err) => { this.emit("launcherror", err); }).bind(this));
            Promise.all([
                this.sendCommand("gdb-set target-async on"),
                this.sendCommand("environment-directory \"" + escape(cwd) + "\""),
                this.sendCommand("target-select remote " + target)
            ]).then(() => {
                this.emit("debug-ready");
                resolve(true); // Dubner added true
            }, reject);
        });
    }
    convert_mi_variables(buf) {
        var retval = buf;
        if (typeof buf == "string") {
            // The cobcd.py extension to gdb returns a string when issued the
            // command "cprint/m ?".  The processing up to this point converts that
            // string to a string, meaning any embedded control characters
            // have been converted to '\"', '\\', '\n,' and so on.
            // I don't know who is doing that, or why.  Rather than dig down,
            // I decided to minimize the impact and just transmogrify it back
            // here.
            // With the advent of using cob_get_field_string() to get the contents
            // of variables, I started seeing GDB messages starting with '=' embedded
            // along with the information we need, which starts with ~" and which
            // has to end with something like 14^done.
            // So, I simply throw away any lines in the multi-line response that
            // start with '='
            //console.log("###1###:" + buf);
            var lines = buf.split('\n');
            var lineno;
            var buf = "";
            for (lineno = 0; lineno < lines.length; lineno++) {
                if (lines[lineno].length > 0 && lines[lineno][0] != '=') {
                    buf += lines[lineno];
                    buf += '\n';
                }
            }
            //console.log("###2###:" + buf);
            const leading_variables = '~"variables=';
            const leading_value = '~"value=';
            if (buf.substr(0, leading_variables.length) == leading_variables || buf.substr(0, leading_value.length) == leading_value) {
                //console.log("convert_mi_variables(1): ", buf); //DUBNERDEBUG
                var n1 = buf.indexOf('\n');
                if (n1 > -1) {
                    var n2 = buf.indexOf('\n', n1 + 1);
                    if (n2 > -1) {
                        // Find the ^done line, and put it in front
                        retval = buf.substr(n1 + 1, n2 - (n1 + 1));
                        retval += ',';
                        retval += buf.substr(2, (n1 - 1) - 2); // Skip the leading '~"' and the trailing '"'
                    }
                    else {
                        // There is no ^done line at this time -- it'll probably be the next line
                        // we get from GDB, but that doesn't do us any good, here -- so, synthesize what
                        // we need to keep going.
                        this.syntheticToken = this.currentToken - 1;
                        retval = (this.currentToken - 1) + "^done,";
                        retval += buf.substr(2, (n1 - 1) - 2); // Skip the leading '~"' and the trailing '"'
                    }
                }
                var temp = retval;
                retval = "";
                for (var i = 0; i < temp.length;) {
                    var ch = temp.charAt(i++);
                    if (ch == "\\") {
                        ch = temp.charAt(i++);
                        if (ch == 'n') {
                            retval += '\n';
                        }
                        else {
                            retval += ch;
                        }
                    }
                    else {
                        retval += ch;
                    }
                }
                temp = retval;
                retval = "";
                for (var i = 0; i < temp.length; i++) {
                    var ch = temp.charAt(i);
                    if (ch == '/') {
                        ch = '/'; // '\u203F'; // undertie
                    }
                    retval += ch;
                }
                //console.log("PROCESSED: ",retval);
                // The reader should be aware that the retval here is based very much on the output
                // of GDB, in machine=interface mode, to a '-stack-list-variables --all-values' command.
                //
                // If a variable's contents require a backslash, the text string must be '\\' at that point.
                // Likewise, '\"' is necessary to produce a single double-quote character.
                //
                // The following statement provides visibility to GDB-Python response
                // to a 'print/m ?' command
                //console.log("convert_mi_variables(2): ", retval); // DUBNERDEBUG
                // The following statements allow for the injection of a hardcoded simulated "response"
                // to such a command.  Figuring out how to format the response wasn't easy.  Nor was the
                // process of debugging the public-domain debugger extension code so that it worked properly;
                // this hard-coded injection proved very useful.
                //
                //				var nfound = retval.indexOf('[');
                //				if(nfound != -1){
                //					var newr = retval.substr(0,nfound)
                //					newr = newr + "[{name=\"L01\",value=\"{_groupvalue=\\\"Snap!\\\",fullname=\\\"RobertDubner\\\",L02={_groupvalue=\\\"Crackle!\\\",firstname = \\\"Robert\\\", lastname = \\\"Dubner\\\",L03={_groupvalue=\\\"Pop!\\\",firstname = \\\"Judy\\\", lastname = \\\"Ruderman\\\"}}}\"}]\n";
                //					newr = newr + "[{name=\"L00L\",value=\"{_groupvalue=\\\"Sn\\\\\\\"ap!\\\",fullname=\\\"Robert\\\\\\\"Dubner\\\",L02={_groupvalue=\\\"Crackle!\\\",firstname = \\\"Robert\\\", lastname = \\\"\\\\\\\"Dubner\\\\\\\"\\\",L03={_groupvalue=\\\"Pop!\\\",firstname = \\\"Judy\\\", lastname = \\\"Ruderman\\\"}}}\"}]\n";
                //					newr = newr + "[{name=\"L01\",value=\"===\\\">\\\\/<\\\"===\"}]\n"; // Produces ===">\/<"===
                //					newr = newr + "[{name=\"L01\",value=\"\\\"===\\\">\\\\/<\\\"===,===\\\">\\\\/<\\\"===\"}]\n"; // Produces ===">\/<"===
                //					// original newr = newr + "[{name="L00",value="\"Standalone\""},{name="slash1",value="\"***\\\"***\", ' ' <repeats 14 times>"},{name="slash2",value="\"\\\"**\\\"***\", ' ' <repeats 14 times>"},{name="slash3",value="\"***\\\"**\\\"\", ' ' <repeats 14 times>"},{name="slash4",value="\"\\\"**\\\"**\\\"\", ' ' <repeats 14 times>"},{name="slash5",value="\"***\\\"***},{***\\\"***    \""},{name="slash6",value="\"\\\"**\\\"***},{\\\"**\\\"***    \""},{name="slash7",value="\"***\\\"**\\\"{,}***\\\"**\\\"    \""},{name="slash8",value="\"\\\"**\\\"**\\\"{,}\\\"**\\\"**\\\"    \""},{name="L01",value="{fullname = \"RobertDubner\", L02 = {firstname = \"Robert\", lastname = \"Dubner\"}}"},{name="L01_2",value="{fullname = \"JudyRuderman\", L02 = {firstname = \"Judy\", lastname = \"Ruderman\"}}"}]""
                //					newr = newr + "[{name=\"L00\",value=\"\\\"Standalone\\\"\"},{name=\"slash1\",value=\"\\\"***\\\\\\\"***\\\", ' ' <repeats 14 times>\"},{name=\"slash2\",value=\"\\\"\\\\\\\"**\\\\\\\"***\\\", ' ' <repeats 14 times>\"},{name=\"slash3\",value=\"\\\"***\\\\\\\"**\\\\\\\"\\\", ' ' <repeats 14 times>\"},{name=\"slash4\",value=\"\\\"\\\\\\\"**\\\\\\\"**\\\\\\\"\\\", ' ' <repeats 14 times>\"},{name=\"slash5\",value=\"\\\"***\\\\\\\"***},{***\\\\\\\"***    \\\"\"},{name=\"slash6\",value=\"\\\"\\\\\\\"**\\\\\\\"***},{\\\\\\\"**\\\\\\\"***    \\\"\"},{name=\"slash7\",value=\"\\\"***\\\\\\\"**\\\\\\\"{,}***\\\\\\\"**\\\\\\\"    \\\"\"},{name=\"slash8\",value=\"\\\"\\\\\\\"**\\\\\\\"**\\\\\\\"{,}\\\\\\\"**\\\\\\\"**\\\\\\\"    \\\"\"},{name=\"L01\",value=\"{fullname = \\\"RobertDubner\\\", L02 = {firstname = \\\"Robert\\\", lastname = \\\"Dubner\\\"}}\"},{name=\"L01_2\",value=\"{fullname = \\\"JudyRuderman\\\", L02 = {firstname = \\\"Judy\\\", lastname = \\\"Ruderman\\\"}}\"}]\n";
                // Original :
                // [{name="L00",value="\"Standalone          \""},{name="slash1",value="\"***\\\"***\""},{name="slash2",value="\"\\\"**\\\"***\""},{name="slash3",value="\"***\\\"**\\\"\""},{name="slash4",value="\"\\\"**\\\"**\\\"\""},{name="slash5",value="\"***\\\"***},{***\\\"***    \""},{name="slash6",value="\"\\\"**\\\"***},{\\\"**\\\"***    \""},{name="slash7",value="\"***\\\"**\\\"{,}***\\\"**\\\"    \""},{name="slash8",value="\"\\\"**\\\"**\\\"{,}\\\"**\\\"**\\\"    \""},{name="L01_1",value="{fullname = \"---\\\\^/---\", L02 = {firstname = \"/---\\\\^/---\\\\         \", lastname = \"\\\\---\\\\^/---/         \"}}"},{name="L01_2",value="{fullname = \"\\\\\\\"---\\\"\\\\^/\\\"---/\\\"     \", L02 = {firstname = \"\\\\---\\\"\\\\^/\\\"---/\\\"      \", lastname = \"\\\\\\\"---\\\"\\\\^/\\\"---/\\\"     \"}}"}]
                // With backslashes and double-quotes escaped:
                // newr = newr + "[{name=\"L00\",value=\"\\\"Standalone          \\\"\"},{name=\"slash1\",value=\"\\\"***\\\\\\\"***\\\"\"},{name=\"slash2\",value=\"\\\"\\\\\\\"**\\\\\\\"***\\\"\"},{name=\"slash3\",value=\"\\\"***\\\\\\\"**\\\\\\\"\\\"\"},{name=\"slash4\",value=\"\\\"\\\\\\\"**\\\\\\\"**\\\\\\\"\\\"\"},{name=\"slash5\",value=\"\\\"***\\\\\\\"***},{***\\\\\\\"***    \\\"\"},{name=\"slash6\",value=\"\\\"\\\\\\\"**\\\\\\\"***},{\\\\\\\"**\\\\\\\"***    \\\"\"},{name=\"slash7\",value=\"\\\"***\\\\\\\"**\\\\\\\"{,}***\\\\\\\"**\\\\\\\"    \\\"\"},{name=\"slash8\",value=\"\\\"\\\\\\\"**\\\\\\\"**\\\\\\\"{,}\\\\\\\"**\\\\\\\"**\\\\\\\"    \\\"\"},{name=\"L01_1\",value=\"{fullname = \\\"---\\\\\\\\^/---\\\", L02 = {firstname = \\\"/---\\\\\\\\^/---\\\\\\\\         \\\", lastname = \\\"\\\\\\\\---\\\\\\\\^/---/         \\\"}}\"},{name=\"L01_2\",value=\"{fullname = \\\"\\\\\\\\\\\\\\\"---\\\\\\\"\\\\\\\\^/\\\\\\\"---/\\\\\\\"     \\\", L02 = {firstname = \\\"\\\\\\\\---\\\\\\\"\\\\\\\\^/\\\\\\\"---/\\\\\\\"      \\\", lastname = \\\"\\\\\\\\\\\\\\\"---\\\\\\\"\\\\\\\\^/\\\\\\\"---/\\\\\\\"     \\\"}}\"}]\n";
                // Original
                // [{name="L01_1",value="{fullname = \"FULLNAME            \", L02 = {firstname = \"FIRSTNAME           \", lastname = \"LASTNAME            \"}}"},{name="L01_2",value="{fullname = \"FULLNAME2           \", L02 = {firstname = \"FIRSTNAME2          \", lastname = \"LASTNAME2           \"}}"}]
                //newr = newr + "[{name=\"L01_1\",value=\"{fullname = \\\"FULLNAME            \\\", L02 = {firstname = \\\"FIRSTNAME           \\\", lastname = \\\"LASTNAME            \\\"}}\"},{name=\"L01_2\",value=\"{fullname = \\\"FULLNAME2           \\\", L02 = {firstname = \\\"FIRSTNAME2          \\\", lastname = \\\"LASTNAME2           \\\"}}\"}]\n";
                //newr = newr + "[{name=\"L00\",value=\"\\\"Standalone          \\\"\"},{name=\"slash1\",value=\"\\\"***\\\\\\\"***\\\"\"},{name=\"slash2\",value=\"\\\"\\\\\\\"**\\\\\\\"***\\\"\"},{name=\"slash3\",value=\"\\\"***\\\\\\\"**\\\\\\\"\\\"\"},{name=\"slash4\",value=\"\\\"\\\\\\\"**\\\\\\\"**\\\\\\\"\\\"\"},{name=\"slash5\",value=\"\\\"***\\\\\\\"***},{***\\\\\\\"***    \\\"\"},{name=\"slash6\",value=\"\\\"\\\\\\\"**\\\\\\\"***},{\\\\\\\"**\\\\\\\"***    \\\"\"},{name=\"slash7\",value=\"\\\"***\\\\\\\"**\\\\\\\"{,}***\\\\\\\"**\\\\\\\"    \\\"\"},{name=\"slash8\",value=\"\\\"\\\\\\\"**\\\\\\\"**\\\\\\\"{,}\\\\\\\"**\\\\\\\"**\\\\\\\"    \\\"\"},{name=\"L01_1\",value=\"{_groupvalue=\\\"He\\\\\\\\\\\\\\\"llo\\\",fullname = \\\"---\\\\\\\\^/---\\\", L02 = {_groupvalue=\\\"\\\\\\\"D\\\\\\\\oo\\\\\\\"\\\\\\\\\\\\\\\"med\\\",firstname = \\\"/---\\\\\\\\^/---\\\\\\\\         \\\", lastname = \\\"\\\\\\\\---\\\\\\\\^/---/         \\\"}}\"},{name=\"L01_2\",value=\"{fullname = \\\"\\\\\\\\\\\\\\\"---\\\\\\\"\\\\\\\\^/\\\\\\\"---/\\\\\\\"     \\\", L02 = {firstname = \\\"\\\\\\\\---\\\\\\\"\\\\\\\\^/\\\\\\\"---/\\\\\\\"      \\\", lastname = \\\"\\\\\\\\\\\\\\\"---\\\\\\\"\\\\\\\\^/\\\\\\\"---/\\\\\\\"     \\\"}}\"}]\n";
                //					retval = newr ;
                //					console.log("convert_mi_variables(3): ", newr); // DUBNERDEBUG
                //				}
            }
        }
        return retval;
    }
    stdout(data) {
        if (this.buffer == undefined) {
            this.buffer = "";
        }
        if (trace)
            this.log("stderr", "stdout: " + data);
        if (typeof data == "string") {
            this.buffer += data;
            //console.log("(A): ", data); //DUBNERDEBUG
        }
        else {
            // DUBNER SEZ:  By replacing "location-reached" with "end-stepping-range", we prevent
            // the alarming red EXCEPTION OCCURRED warning when we replace "-exec-next" with "cnext"
            this.buffer += data.toString("utf8").replace("location-reached", "end-stepping-range");
            //console.log("(B): ", data.toString("utf8"));  //DUBNERDEBUG
        }
        var end = this.buffer.lastIndexOf('\n');
        // To avoid spurious, albeit harmless, error messages, we
        // need to avoid processing "123^done" messages when
        // we already generated a synthetic "123^done message in convert_mi_variables"
        if (end != -1) {
            var frag = this.buffer.slice(0, end);
            var patt = new RegExp("^[0-9]+[\^]done");
            if (patt.test(frag)) {
                var token = parseInt(frag);
                if (token <= this.syntheticToken) {
                    // We generated this token number, which means that
                    // this ^done, if processed, will confuse VSC, because it
                    // doesn't expect to see tokens twice.
                    this.buffer = "";
                    end = -1;
                }
            }
        }
        // We need to know if the program is *stopped or *running.  This is
        // important, for example, in sendCommand().  So, we need to pull
        // this.buffer apart so that we can look at the pieces.
        // (Yes, I know I am doing this at least twice.  There's a lot of asynchronous stuff
        // going on here, and I am hacking at it.  I'm lucky it works at all; I really don't
        // want to try to refactor it when I frankly am not sure what it is doing, or even what
        // it is *supposed* to be doing, and how it's trying to do it.)
        const leading_running = '*running';
        const leading_stopped = '*stopped';
        var lines = this.buffer.split('\n');
        for (var i = 0; i < lines.length; i++) {
            // Don't break out of this loop. When stepping it's common for *running and *stopped
            // to appear in the response.
            var line = lines[i];
            if (line.substr(0, leading_running.length) == leading_running) {
                this.starState = leading_running;
            }
            if (line.substr(0, leading_stopped.length) == leading_stopped) {
                this.starState = leading_stopped;
            }
        }
        if (end != -1) {
            this.buffer = this.convert_mi_variables(this.buffer);
            // We need to find end again, in case convert_mi_variables
            // changed the length.
            end = this.buffer.lastIndexOf('\n');
            this.onOutput(this.buffer.substr(0, end));
            this.buffer = this.buffer.substr(end + 1);
        }
        if (this.buffer.length) {
            if (this.onOutputPartial(this.buffer)) {
                this.buffer = "";
            }
        }
    }
    stderr(data) {
        if (typeof data == "string")
            this.errbuf += data;
        else
            this.errbuf += data.toString("utf8");
        const end = this.errbuf.lastIndexOf('\n');
        if (end != -1) {
            this.onOutputStderr(this.errbuf.substr(0, end));
            this.errbuf = this.errbuf.substr(end + 1);
        }
        if (this.errbuf.length) {
            this.logNoNewLine("stderr", this.errbuf);
            this.errbuf = "";
        }
    }
    onOutputStderr(lines) {
        lines = lines.split('\n');
        lines.forEach(line => {
            this.log("stderr", line);
        });
    }
    onOutputPartial(line) {
        if (couldBeOutput(line)) {
            this.logNoNewLine("stdout", line);
            return true;
        }
        return false;
    }
    onOutput(lines) {
        lines = lines.split('\n');
        lines.forEach(line => {
            if (couldBeOutput(line)) {
                if (!gdbMatch.exec(line))
                    this.log("stdout", line);
            }
            else {
                const parsed = mi_parse_1.parseMI(line);
                if (this.debugOutput)
                    this.log("log", "GDB -> App: " + JSON.stringify(parsed));
                let handled = false;
                if (parsed.token !== undefined) {
                    if (this.handlers[parsed.token]) {
                        this.handlers[parsed.token](parsed);
                        delete this.handlers[parsed.token];
                        handled = true;
                    }
                }
                if (!handled && parsed.resultRecords && parsed.resultRecords.resultClass == "error") {
                    this.log("stderr", parsed.result("msg") || line);
                }
                if (parsed.outOfBandRecord) {
                    parsed.outOfBandRecord.forEach(record => {
                        if (record.isStream) {
                            this.log(record.type, record.content);
                        }
                        else {
                            if (record.type == "exec") {
                                this.emit("exec-async-output", parsed);
                                if (record.asyncClass == "running")
                                    this.emit("running", parsed);
                                else if (record.asyncClass == "stopped") {
                                    const reason = parsed.record("reason");
                                    if (trace)
                                        this.log("stderr", "stop: " + reason);
                                    if (reason == "breakpoint-hit")
                                        this.emit("breakpoint", parsed);
                                    else if (reason == "end-stepping-range")
                                        this.emit("step-end", parsed);
                                    else if (reason == "function-finished")
                                        this.emit("step-out-end", parsed);
                                    else if (reason == "signal-received")
                                        this.emit("signal-stop", parsed);
                                    else if (reason == "exited-normally")
                                        this.emit("exited-normally", parsed);
                                    else if (reason == "exited") { // exit with error code != 0
                                        this.log("stderr", "Program exited with code " + parsed.record("exit-code"));
                                        this.emit("exited-normally", parsed);
                                    }
                                    else {
                                        this.log("console", "Not implemented stop reason (assuming exception): " + reason);
                                        this.emit("stopped", parsed);
                                    }
                                }
                                else
                                    this.log("log", JSON.stringify(parsed));
                            }
                            else if (record.type == "notify") {
                                if (record.asyncClass == "thread-created") {
                                    this.emit("thread-created", parsed);
                                }
                                else if (record.asyncClass == "thread-exited") {
                                    this.emit("thread-exited", parsed);
                                }
                            }
                        }
                    });
                    handled = true;
                }
                if (parsed.token == undefined && parsed.resultRecords == undefined && parsed.outOfBandRecord.length == 0)
                    handled = true;
                if (!handled)
                    this.log("log", "Unhandled: " + JSON.stringify(parsed));
            }
        });
    }
    start() {
        return new Promise((resolve, reject) => {
            this.once("ui-break-done", () => {
                this.log("console", "Running executable");
                this.sendCommand("exec-run").then((info) => {
                    if (info.resultRecords.resultClass == "running")
                        resolve(true); // Dubner added true
                    else
                        reject();
                }, reject);
            });
        });
    }
    stop() {
        if (this.isSSH) {
            const proc = this.stream;
            const to = setTimeout(() => {
                proc.signal("KILL");
            }, 1000);
            this.stream.on("exit", function (code) {
                clearTimeout(to);
            });
            this.sendRaw("-gdb-exit");
        }
        else {
            const proc = this.process;
            const to = setTimeout(() => {
                process.kill(-proc.pid);
            }, 1000);
            this.process.on("exit", function (code) {
                clearTimeout(to);
            });
            this.sendRaw("-gdb-exit");
        }
    }
    detach() {
        const proc = this.process;
        const to = setTimeout(() => {
            process.kill(-proc.pid);
        }, 1000);
        this.process.on("exit", function (code) {
            clearTimeout(to);
        });
        this.sendRaw("-target-detach");
    }
    interrupt() {
        if (trace)
            this.log("stderr", "interrupt");
        return new Promise((resolve, reject) => {
            this.sendCommand("exec-interrupt").then((info) => {
                resolve(info.resultRecords.resultClass == "done");
            }, reject);
        });
    }
    continue(reverse = false) {
        if (trace)
            this.log("stderr", "continue");
        return new Promise((resolve, reject) => {
            this.sendCommand("exec-continue" + (reverse ? " --reverse" : "")).then((info) => {
                resolve(info.resultRecords.resultClass == "running");
            }, reject);
        });
    }
    next(reverse = false) {
        if (trace)
            this.log("stderr", "next");
        return new Promise((resolve, reject) => {
            //			this.sendCommand("exec-next" + (reverse ? " --reverse" : "")).then((info) => {
            this.sendCommand("interpreter-exec --thread 1 --frame 0 console \"cnext\"" + (reverse ? " --reverse" : "")).then((info) => {
                resolve(info.resultRecords.resultClass == "running");
            }, reject);
        });
    }
    step(reverse = false) {
        if (trace)
            this.log("stderr", "step");
        return new Promise((resolve, reject) => {
            this.sendCommand("exec-step" + (reverse ? " --reverse" : "")).then((info) => {
                resolve(info.resultRecords.resultClass == "running");
            }, reject);
        });
    }
    stepOut(reverse = false) {
        if (trace)
            this.log("stderr", "stepOut");
        return new Promise((resolve, reject) => {
            this.sendCommand("exec-finish" + (reverse ? " --reverse" : "")).then((info) => {
                resolve(info.resultRecords.resultClass == "running");
            }, reject);
        });
    }
    goto(filename, line) {
        if (trace)
            this.log("stderr", "goto");
        return new Promise((resolve, reject) => {
            const target = '"' + (filename ? escape(filename) + ":" : "") + line + '"';
            this.sendCommand("break-insert -t " + target).then(() => {
                this.sendCommand("exec-jump " + target).then((info) => {
                    resolve(info.resultRecords.resultClass == "running");
                }, reject);
            }, reject);
        });
    }
    changeVariableOriginal(name, rawValue) {
        if (trace)
            this.log("stderr", "changeVariable");
        return this.sendCommand("gdb-set var " + name + "=" + rawValue);
    }
    changeVariable(name, rawValue) {
        if (trace)
            this.log("stderr", "changeVariable");
        return this.sendStraightCommand("cprint " + name + "=" + rawValue);
    }
    loadBreakPoints(breakpoints) {
        if (trace)
            this.log("stderr", "loadBreakPoints");
        const promisses = [];
        breakpoints.forEach(breakpoint => {
            promisses.push(this.addBreakPoint(breakpoint));
        });
        return Promise.all(promisses);
    }
    setBreakPointCondition(bkptNum, condition) {
        if (trace)
            this.log("stderr", "setBreakPointCondition");
        return this.sendCommand("break-condition " + bkptNum + " " + condition);
    }
    addBreakPoint(breakpoint) {
        if (trace)
            this.log("stderr", "addBreakPoint");
        return new Promise((resolve, reject) => {
            if (this.breakpoints.has(breakpoint))
                return resolve([false, undefined]);
            let location = "";
            if (breakpoint.countCondition) {
                if (breakpoint.countCondition[0] == ">")
                    location += "-i " + numRegex.exec(breakpoint.countCondition.substr(1))[0] + " ";
                else {
                    const match = numRegex.exec(breakpoint.countCondition)[0];
                    if (match.length != breakpoint.countCondition.length) {
                        this.log("stderr", "Unsupported break count expression: '" + breakpoint.countCondition + "'. Only supports 'X' for breaking once after X times or '>X' for ignoring the first X breaks");
                        location += "-t ";
                    }
                    else if (parseInt(match) != 0)
                        location += "-t -i " + parseInt(match) + " ";
                }
            }
            if (breakpoint.raw)
                location += '"' + escape(breakpoint.raw) + '"';
            else
                location += '"' + escape(breakpoint.file) + ":" + breakpoint.line + '"';
            this.sendCommand("break-insert -f " + location).then((result) => {
                if (result.resultRecords.resultClass == "done") {
                    const bkptNum = parseInt(result.result("bkpt.number"));
                    const newBrk = {
                        file: result.result("bkpt.file"),
                        line: parseInt(result.result("bkpt.line")),
                        condition: breakpoint.condition
                    };
                    if (breakpoint.condition) {
                        this.setBreakPointCondition(bkptNum, breakpoint.condition).then((result) => {
                            if (result.resultRecords.resultClass == "done") {
                                this.breakpoints.set(newBrk, bkptNum);
                                resolve([true, newBrk]);
                            }
                            else {
                                resolve([false, undefined]);
                            }
                        }, reject);
                    }
                    else {
                        this.breakpoints.set(newBrk, bkptNum);
                        resolve([true, newBrk]);
                    }
                }
                else {
                    reject(result);
                }
            }, reject);
        });
    }
    removeBreakPoint(breakpoint) {
        if (trace)
            this.log("stderr", "removeBreakPoint");
        return new Promise((resolve, reject) => {
            if (!this.breakpoints.has(breakpoint))
                return resolve(false);
            this.sendCommand("break-delete " + this.breakpoints.get(breakpoint)).then((result) => {
                if (result.resultRecords.resultClass == "done") {
                    this.breakpoints.delete(breakpoint);
                    resolve(true);
                }
                else
                    resolve(false);
            });
        });
    }
    clearBreakPoints() {
        if (trace)
            this.log("stderr", "clearBreakPoints");
        return new Promise((resolve, reject) => {
            this.sendCommand("break-delete").then((result) => {
                if (result.resultRecords.resultClass == "done") {
                    this.breakpoints.clear();
                    resolve(true);
                }
                else
                    resolve(false);
            }, () => {
                resolve(false);
            });
        });
    }
    getThreads() {
        return __awaiter(this, void 0, void 0, function* () {
            if (trace)
                this.log("stderr", "getThreads");
            const ret = [];
            const command = "thread-info";
            const result = yield this.sendCommand(command);
            const threads = result.result("threads");
            return threads.map(element => {
                const ret = {
                    id: parseInt(mi_parse_1.MINode.valueOf(element, "id")),
                    targetId: mi_parse_1.MINode.valueOf(element, "target-id")
                };
                const name = mi_parse_1.MINode.valueOf(element, "name");
                if (name) {
                    ret.name = name;
                }
                return ret;
            });
        });
    }
    getStack(maxLevels, thread) {
        return __awaiter(this, void 0, void 0, function* () {
            if (trace)
                this.log("stderr", "getStack");
            let command = "stack-list-frames";
            if (thread != 0) {
                command += ` --thread ${thread}`;
            }
            if (maxLevels) {
                command += " 0 " + maxLevels;
            }
            const result = yield this.sendCommand(command);
            const stack = result.result("stack");
            const ret = [];
            return stack.map(element => {
                const level = mi_parse_1.MINode.valueOf(element, "@frame.level");
                const addr = mi_parse_1.MINode.valueOf(element, "@frame.addr");
                const func = mi_parse_1.MINode.valueOf(element, "@frame.func");
                const filename = mi_parse_1.MINode.valueOf(element, "@frame.file");
                let file = mi_parse_1.MINode.valueOf(element, "@frame.fullname");
                if (file) {
                    if (this.isSSH)
                        file = path_1.posix.normalize(file);
                    else
                        file = nativePath.normalize(file);
                }
                let line = 0;
                const lnstr = mi_parse_1.MINode.valueOf(element, "@frame.line");
                if (lnstr)
                    line = parseInt(lnstr);
                const from = parseInt(mi_parse_1.MINode.valueOf(element, "@frame.from"));
                return {
                    address: addr,
                    fileName: filename,
                    file: file,
                    function: func || from,
                    level: level,
                    line: line
                };
            });
        });
    }
    getStackVariablesOriginal(thread, frame) {
        return __awaiter(this, void 0, void 0, function* () {
            if (trace)
                this.log("stderr", "getStackVariables");
            const result = yield this.sendCommand(`stack-list-variables --thread ${thread} --frame ${frame} --simple-values`);
            const variables = result.result("variables");
            const ret = [];
            for (const element of variables) {
                const key = mi_parse_1.MINode.valueOf(element, "name");
                const value = mi_parse_1.MINode.valueOf(element, "value");
                const type = mi_parse_1.MINode.valueOf(element, "type");
                ret.push({
                    name: key,
                    valueStr: value,
                    type: type,
                    raw: element
                });
            }
            return ret;
        });
    }
    getStackVariables(thread, frame) {
        return __awaiter(this, void 0, void 0, function* () {
            if (trace)
                this.log("stderr", "getStackVariables");
            //			const result = await this.sendStraightCommand(`cprint/m ?`);
            const result = yield this.sendCommand("interpreter-exec console \"cprint/m ?\"", false);
            const variables = result.result("variables");
            const ret = [];
            for (const element of variables) {
                const key = mi_parse_1.MINode.valueOf(element, "name");
                const value = mi_parse_1.MINode.valueOf(element, "value");
                const type = mi_parse_1.MINode.valueOf(element, "type");
                ret.push({
                    name: key,
                    valueStr: value,
                    type: type,
                    raw: element
                });
            }
            return ret;
        });
    }
    examineMemory(from, length) {
        7;
        if (trace)
            this.log("stderr", "examineMemory");
        return new Promise((resolve, reject) => {
            this.sendCommand("data-read-memory-bytes 0x" + from.toString(16) + " " + length).then((result) => {
                resolve(result.result("memory[0].contents"));
            }, reject);
        });
    }
    evalExpression(name, thread, frame) {
        return __awaiter(this, void 0, void 0, function* () {
            if (trace) {
                this.log("stderr", "evalExpression");
            }
            let command = "interpreter-exec console \"cprint/m ";
            command += name;
            command += "\"";
            //console.log("evalExpression(): ", command); // DUBNERDEBUG
            return yield this.sendCommand(command, false);
            //		let command = "data-evaluate-expression ";
            //		if (thread != 0) {
            //			command += `--thread ${thread} --frame ${frame} `;
            //		}
            //		command += name;
            //		//console.log(command); // DUBNERDEBUG
            //		return await this.sendCommand(command);
        });
    }
    varCreate(expression, name = "-") {
        return __awaiter(this, void 0, void 0, function* () {
            if (trace)
                this.log("stderr", "varCreate");
            const res = yield this.sendCommand(`var-create ${name} @ "${expression}"`);
            return new backend_1.VariableObject(res.result(""));
        });
    }
    varEvalExpression(name) {
        return __awaiter(this, void 0, void 0, function* () {
            if (trace)
                this.log("stderr", "varEvalExpression");
            return this.sendCommand(`var-evaluate-expression ${name}`);
        });
    }
    varListChildren(name) {
        return __awaiter(this, void 0, void 0, function* () {
            if (trace)
                this.log("stderr", "varListChildren");
            //TODO: add `from` and `to` arguments
            const res = yield this.sendCommand(`var-list-children --all-values ${name}`);
            const children = res.result("children") || [];
            const omg = children.map(child => new backend_1.VariableObject(child[1]));
            return omg;
        });
    }
    varUpdate(name = "*") {
        return __awaiter(this, void 0, void 0, function* () {
            if (trace)
                this.log("stderr", "varUpdate");
            return this.sendCommand(`var-update --all-values ${name}`);
        });
    }
    varAssign(name, rawValue) {
        return __awaiter(this, void 0, void 0, function* () {
            if (trace)
                this.log("stderr", "varAssign");
            return this.sendCommand(`var-assign ${name} ${rawValue}`);
        });
    }
    logNoNewLine(type, msg) {
        this.emit("msg", type, msg);
    }
    log(type, msg) {
        this.emit("msg", type, msg[msg.length - 1] == '\n' ? msg : (msg + "\n"));
    }
    sendUserInput(command, threadId = 0, frameLevel = 0) {
        if (command.startsWith("-")) {
            return this.sendCommand(command.substr(1));
        }
        else {
            return this.sendCliCommand(command, threadId, frameLevel);
        }
    }
    sendRaw(raw) {
        console.log("sendRaw():", raw, '(', this.starState, ')'); // DUBNERDEBUG
        if (this.printCalls)
            this.log("log", raw);
        if (this.isSSH)
            this.stream.write(raw + "\n");
        else
            this.process.stdin.write(raw + "\n");
    }
    sendCliCommand(command, threadId = 0, frameLevel = 0) {
        return __awaiter(this, void 0, void 0, function* () {
            // For some reason, somebody, somewhere, is sending a console command "process.pid", which
            // just confuses everybody and creates an error message the user can see.
            // Let's nip this one in the bud.
            if (command == "process.pid") {
                return;
            }
            let miCommand = "interpreter-exec ";
            if (threadId != 0) {
                miCommand += `--thread ${threadId} --frame ${frameLevel} `;
            }
            miCommand += `console "${command.replace(/[\\"']/g, '\\$&')}"`;
            yield this.sendCommand(miCommand);
        });
    }
    sendCommand(command, suppressFailure = false) {
        //console.log("(C)", command);
        const sel = this.currentToken++;
        return new Promise((resolve, reject) => {
            this.handlers[sel] = (node) => {
                if (node && node.resultRecords && node.resultRecords.resultClass === "error") {
                    if (suppressFailure) {
                        this.log("stderr", `WARNING: Error executing command '${command}'`);
                        resolve(node);
                    }
                    else
                        reject(new backend_1.MIError(node.result("msg") || "Internal error", command));
                }
                else
                    resolve(node);
            };
            if (this.starState == "*running") {
                // When in running mode, assume the command is for the inferior
                // rather than for gdb.  We will pay attention only to an instruction
                // encased in 'interpreter-exec', and we will extract the meat of
                // that instruction and send it unadorned.
                if (command.indexOf('interpreter-exec') == 0) {
                    var n1 = command.indexOf('"');
                    var n2 = command.lastIndexOf('"');
                    if (n1 != -1 && n2 != -1) {
                        command = command.slice(n1 + 1, n2);
                        this.sendRaw(command);
                        // Welcome to the world of black magic.  On a number of occasions, I observed
                        // that entering 12345 to an ACCEPT statement resulted in 02345, or 00345.  I
                        // have been unable to figure out why.  But after adding a "pause", I never saw
                        // the problem.  Very possibly coincidence.  Real-time bugs don't get any easier
                        // to find, no matter how old I get.  But until it crops up again, I'm leaving
                        // this pause in.
                        pausecomp(100);
                        this.sendRaw("\n");
                    }
                }
            }
            else {
                this.sendRaw(sel + "-" + command);
            }
        });
    }
    sendStraightCommand(command, suppressFailure = false) {
        const sel = this.currentToken++;
        return new Promise((resolve, reject) => {
            this.handlers[sel] = (node) => {
                if (node && node.resultRecords && node.resultRecords.resultClass === "error") {
                    if (suppressFailure) {
                        this.log("stderr", `WARNING: Error executing command '${command}'`);
                        resolve(node);
                    }
                    else
                        reject(new backend_1.MIError(node.result("msg") || "Internal error", command));
                }
                else
                    resolve(node);
            };
            //console.log("<command> " + sel + command); // DUBNERDEBUG
            this.sendRaw(sel + command);
        });
    }
    isReady() {
        return this.isSSH ? this.sshReady : !!this.process;
    }
}
exports.MI2 = MI2;
//# sourceMappingURL=mi2.js.map